package kms

import (
	"errors"
	"fmt"
	"sync"

	log "github.com/sirupsen/logrus"

	"code.fbi.h-da.de/danet/proto-kms/quantumlayer"
)

type Status int

const (
	RESERVED = iota
	AVAILABLE
	UNKNOWN
	USED
)

type kmsKS interface {
	KeyChopper(bulkKey *quantumlayer.QuantumLayerBulkKey) (err error)
	addKey(int64, [8]byte)
}

// holds a single ready to bit key, length can be configured
type kmsKSElement struct {
	keyID  uint64
	key    []byte // a key
	status Status
}

type kmsKeyStore struct {
	keyStoreMutex    sync.Mutex
	underlyingBulkId int64
	keyStore         map[uint64]*kmsKSElement
	keySingleSize    uint // the size of a single key, given as unit of bits
	indexCounter     uint64
}

func NewKmsKeyStore(desiredkeySingleSizeLength uint) *kmsKeyStore {
	return &kmsKeyStore{
		underlyingBulkId: 0,
		keyStore:         make(map[uint64]*kmsKSElement),
		keySingleSize:    desiredkeySingleSizeLength,
		indexCounter:     0,
	}
}

func (ks *kmsKeyStore) addKey(keyId uint64, keyToadd []byte) {
	ks.keyStoreMutex.Lock()
	defer ks.keyStoreMutex.Unlock()

	// test for collisions
	if _, notThere := ks.keyStore[keyId]; notThere {
		log.Errorf("Whop: addKey collission of key id %s for bulkKeyID %d", keyId, ks.underlyingBulkId)
		return
	}

	newKeyElement := &kmsKSElement{
		keyID:  keyId,
		key:    keyToadd,
		status: AVAILABLE,
	}
	// ok to add
	ks.keyStore[newKeyElement.keyID] = newKeyElement
}

func (ks *kmsKeyStore) GetKey() (*kmsKSElement, error) {
	ks.keyStoreMutex.Lock()
	defer ks.keyStoreMutex.Unlock()

	key, ok := ks.keyStore[ks.indexCounter]
	if !ok {
		return nil, fmt.Errorf("Could not find a key for id: %d", ks.indexCounter)
	}
	if key.status == AVAILABLE {
		// change status of key to reserved
		key.status = RESERVED
		// increase index counter
		ks.indexCounter = ks.indexCounter + 1

		return key, nil
	}
	return nil, fmt.Errorf("Key with index: %d is already reserved.", ks.indexCounter)
}

// Takes a bulk of keys and chops them in chopFactor keys each
// Any remainder is discarded
func (ks *kmsKeyStore) KeyChopper(bulkKey *quantumlayer.QuantumLayerBulkKey) (err error) {
	if ks.keySingleSize == 0 {
		err = errors.New("KeyChopper: no  keySingleSize set")
		return err
	}
	// TODO check if multiple of 8 (1 Byte)

	if bulkKey.BulkKeyLength != len(*bulkKey.BulkKey) {
		err = errors.New("bulkKey length mismatch")
		return err
	}

	// Store underlying bulkID
	ks.underlyingBulkId = bulkKey.BulkKeyId

	var keyId uint64
	keyId = 0

	// Let's chop!
	chopFactor := ks.keySingleSize >> 3
	key := *bulkKey.BulkKey
	for len(key) > int(chopFactor) {
		tmpkey := key[:chopFactor]
		ks.addKey(keyId, tmpkey)
		keyId++
		// shorten the key storage
		key = key[chopFactor:]
	}
	return nil
}
