package kms

import (
	"context"
	"encoding/base64"
	"flag"
	"fmt"
	"net"
	"time"

	"github.com/google/uuid"
	log "github.com/sirupsen/logrus"

	pb "code.fbi.h-da.de/danet/proto-kms/api/gen/proto/go/kmsintercom"
	"google.golang.org/grpc"
	"google.golang.org/grpc/codes"
	"google.golang.org/grpc/health"
	healthpb "google.golang.org/grpc/health/grpc_health_v1"
	"google.golang.org/grpc/status"
)

type kmsTalkerServer struct {
	pb.UnimplementedKmsTalkerServer
	keyNegotiationMap map[uuid.UUID]*kmsKSElement
	eKMS              *EKMS
}

// This must somehow find out and agree to a specific key lenght
func (s *kmsTalkerServer) InterComCapabilities(ctx context.Context, in *pb.InterComCapabilitiesRequest) (capReply *pb.InterComCapabilitiesReply, err error) {
	log.Debugf("Received: %v", in.GetMyKmsName())

	// TODO: Call to ksp := NewKmsKeyStore(<desired-size-of-each-key-in-bits)
	// this to be stored in the serving QLE QuantumElement struct under keyStorePeer
	// Further, the KMS peers have to agree on a ready-to-be-used keyBulk based on the bulkId
	// This requires to go through the rawBulkKeys of type QuantumElement and lookup a bulkId both side do know Once agreed upon on keyBulk, this here has to call KeyChopper of the actual NewKmsKeyStore

	return &pb.InterComCapabilitiesReply{
		PeerKmsName: "whatever",
	}, nil
}

func (s *kmsTalkerServer) SyncQkdBulk(ctx context.Context, in *pb.SyncQkdBulkRequest) (*pb.SyncQkdBulkResponse, error) {
	// check if a peer exists
	peer, ok := s.eKMS.KmsPeers[in.GetInterComAddr()]
	if !ok {
		//TODO: proper error message
		return nil, status.Errorf(codes.Internal, "peer does not exist")
	}

	for _, bulkId := range in.BulkId {
		rawBulkKeyIds := keysOfMap[int64](peer.servingQLE.rawBulkKeys)
		log.Info("PEER BULK KEYS: ", rawBulkKeyIds)
		if bulk, ok := peer.servingQLE.rawBulkKeys[bulkId]; ok {
			peer.servingQLE.keyStorePeer = NewKmsKeyStore(256)
			if err := peer.servingQLE.keyStorePeer.KeyChopper(bulk); err != nil {
				//TODO: proper error message
				return nil, status.Errorf(codes.Internal, "chopping failed")
			}

			return &pb.SyncQkdBulkResponse{
				Timestamp: time.Now().Unix(),
				BulkId:    bulkId,
			}, nil
		}
	}

	return nil, status.Errorf(codes.Unimplemented, "method SyncQkdBulk not implemented")
}

func (s *kmsTalkerServer) InterComTransportKeyNegotiation(ctx context.Context, in *pb.InterComTransportKeyNegotiationRequest) (capReply *pb.InterComTransportKeyNegotiationResponse, err error) {
	// NOTE: For the current prototype it is assumed that a negotiation request
	// is always valid. In the future an incoming negotiation request should
	// also check for a suitable forwarding assignment from the controller.
	keyID := in.KeyToUse
	pathId, err := uuid.Parse(in.PathID)
	if err != nil {
		return nil, status.Errorf(codes.InvalidArgument, "")
	}

	route, ok := s.eKMS.routingTable[pathId]
	if !ok {
		return nil, status.Errorf(codes.Internal, "There is no route for the given pathID: %s .", in.PathID)
	}

	//TODO: This limits a pathId so that it is only possible to send one single
	//payload.
	//if _, ok := s.keyNegotiationMap[pathId]; ok {
	//	return nil, status.Errorf(codes.Internal, "A transport key for pathID: %s has already been negotiated.", in.PathID)
	//}

	quantumElementRemoteKeyStore := route.Previous.servingQLE.keyStorePeer

	key, ok := quantumElementRemoteKeyStore.keyStore[keyID]
	if !ok {
		return nil, status.Errorf(codes.Internal, "Key with ID: %s could not be found.", keyID)
	}

	s.keyNegotiationMap[pathId] = key

	return &pb.InterComTransportKeyNegotiationResponse{Timestamp: time.Now().Unix()}, nil
}

func (s *kmsTalkerServer) KeyForwarding(ctx context.Context, in *pb.KeyForwardingRequest) (capReply *pb.KeyForwardingResponse, err error) {
	pathId, err := uuid.Parse(in.PathId)
	if err != nil {
		return nil, status.Errorf(codes.InvalidArgument, "")
	}

	decryptKey, ok := s.keyNegotiationMap[pathId]
	if !ok {
		return nil, status.Errorf(codes.InvalidArgument, "")
	}

	route, ok := s.eKMS.routingTable[pathId]
	if !ok {
		return nil, status.Errorf(codes.Internal, "There is no route for the given pathID: %s .", in.PathId)
	}

	payloadAsByte, err := base64.StdEncoding.DecodeString(in.Payload)
	if err != nil {
		return nil, status.Errorf(codes.Internal, "%s", err)
	}

	decryptedPayload, err := route.Previous.et.Decrypt(payloadAsByte, decryptKey.key)
	if err != nil {
		return nil, status.Errorf(codes.Internal, "%s", err)
	}

	log.Infof("The eKMS: %s received the Payload: %s", s.eKMS.kmsName, decryptedPayload)

	if route.Next != nil {
		log.Info("SEEEENDING: ", s.eKMS.kmsName)
		go route.Next.SendPayload(decryptedPayload, pathId)
	} else {
		fmt.Println("THE DECRYPTED PAYLOAD: ", string(decryptedPayload))
	}

	return &pb.KeyForwardingResponse{Timestamp: time.Now().Unix()}, nil
}

func StartInterComm(addr string, eKMS *EKMS) {
	flag.Parse()

	lis, err := net.Listen("tcp", addr)
	if err != nil {
		log.Fatalf("failed to listen: %v", err)
	}
	s := grpc.NewServer()
	healthCheck := health.NewServer()

	healthpb.RegisterHealthServer(s, healthCheck)
	pb.RegisterKmsTalkerServer(s, &kmsTalkerServer{
		keyNegotiationMap: make(map[uuid.UUID]*kmsKSElement),
		eKMS:              eKMS,
	})

	go func() {
		// set status to serving
		healthCheck.SetServingStatus(pb.KmsTalker_ServiceDesc.ServiceName, healthpb.HealthCheckResponse_SERVING)
		// TODO: add logic for adjusting health status based on operating status of
		// the services
		//    for{}
	}()

	log.Infof("server listening at %v", lis.Addr())
	if err := s.Serve(lis); err != nil {
		log.Fatalf("failed to serve: %v", err)
	}
}
